/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/* gpdLib2.c*/
/*
 * @(#)gpdLib2.c	1.6 1/13/94 
 *
 *	Authors:  Marty Kraimer and Ben Chin Cha
 *      Current Author: Claude Saunders
 *      Date:	07-13-93
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 */

#ifdef vxWorks
#include <memLib.h>
#include <ctype.h>
#include <stdioLib.h>
#else
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#endif

#include        "gpdLib2.h"
#include	<string.h>

#define 	HASH_NO	256		/* number of hash table entries */

/*The hash algorithm is the algorithm described in			*/
/* Fast Hashing of Variable Length Text Strings, Peter K. Pearson,	*/
/* Communications of the ACM, June 1990					*/

static unsigned char T0[256] = {
 39,159,180,252, 71,  6, 13,164,232, 35,226,155, 98,120,154, 69,
157, 24,137, 29,147, 78,121, 85,112,  8,248,130, 55,117,190,160,
176,131,228, 64,211,106, 38, 27,140, 30, 88,210,227,104, 84, 77,
 75,107,169,138,195,184, 70, 90, 61,166,  7,244,165,108,219, 51,
  9,139,209, 40, 31,202, 58,179,116, 33,207,146, 76, 60,242,124,
254,197, 80,167,153,145,129,233,132, 48,246, 86,156,177, 36,187,
 45,  1, 96, 18, 19, 62,185,234, 99, 16,218, 95,128,224,123,253,
 42,109,  4,247, 72,  5,151,136,  0,152,148,127,204,133, 17, 14,
182,217, 54,199,119,174, 82, 57,215, 41,114,208,206,110,239, 23,
189, 15,  3, 22,188, 79,113,172, 28,  2,222, 21,251,225,237,105,
102, 32, 56,181,126, 83,230, 53,158, 52, 59,213,118,100, 67,142,
220,170,144,115,205, 26,125,168,249, 66,175, 97,255, 92,229, 91,
214,236,178,243, 46, 44,201,250,135,186,150,221,163,216,162, 43,
 11,101, 34, 37,194, 25, 50, 12, 87,198,173,240,193,171,143,231,
111,141,191,103, 74,245,223, 20,161,235,122, 63, 89,149, 73,238,
134, 68, 93,183,241, 81,196, 49,192, 65,212, 94,203, 10,200, 47 
};

static unsigned char hash( char *pname)
{
    unsigned char  h=0;

    while(*pname) {
	h = T0[h^*pname];
	pname++;
    }
    return(h);
  }

void gpdInitPvt(void **pgpdPvt)
{
    ELLLIST        **pgpd;

    pgpd = (ELLLIST **)calloc(HASH_NO, sizeof(ELLLIST *));
    *pgpdPvt = (void *)pgpd;
    return;
}
	
GPDENTRY *gpdFind(void *gpdPvt,char *name,char *namepvt)
{
    unsigned short  hashInd;
    ELLLIST          **pgpd = (ELLLIST **) gpdPvt;
    ELLLIST           *gpdlist;
    GPDENTRY       *pgpdNode;
    
    if (gpdPvt == NULL)
      return(NULL);
    hashInd = (unsigned short)hash(name);
    if ((gpdlist=pgpd[hashInd]) == NULL) return (NULL);
    pgpdNode = (GPDENTRY *) ellFirst(gpdlist);
    while(pgpdNode) {
	if (strcmp(name,(char *)pgpdNode->name) == 0) {
	  if (strlen(namepvt) == 0)
	    return(pgpdNode);
	  else if(strcmp(namepvt,pgpdNode->namepvt)==0)
	    return(pgpdNode);
	}
	pgpdNode = (GPDENTRY *) ellNext((ELLNODE*)pgpdNode);
    }
    return (NULL);
}

GPDENTRY *gpdAdd(void *gpdPvt,char *name,char *namepvt)
{
    unsigned short  hashInd;
    ELLLIST        **pgpd = (ELLLIST **) gpdPvt;
    ELLLIST	*plist;
    GPDENTRY    *pgpdNode;
    
    if (gpdPvt == NULL)
      return(NULL);
    hashInd = (unsigned short)hash(name);
    if (pgpd[hashInd] == NULL) {
	pgpd[hashInd] = (ELLLIST *)calloc(1, sizeof(ELLLIST));
	ellInit(pgpd[hashInd]);
    }
    plist=pgpd[hashInd];
    pgpdNode = (GPDENTRY *) ellFirst(plist);
    while(pgpdNode) {
	if (strcmp(name,(char *)pgpdNode->name) == 0) {
	  if (strlen(namepvt) == 0)
	    return(NULL);
	  else if (strcmp(namepvt,pgpdNode->namepvt)==0)
	    return(NULL);
	}
	pgpdNode = (GPDENTRY *) ellNext((ELLNODE*)pgpdNode);
    }
    pgpdNode = (GPDENTRY *)calloc(1, (unsigned) sizeof(GPDENTRY));
    strcpy(pgpdNode->name,name);
    strcpy(pgpdNode->namepvt,namepvt);
    ellAdd(plist, (ELLNODE*)pgpdNode);
    return (pgpdNode);
}

void gpdDelete(void *gpdPvt,char *name,char *namepvt, void (*userfree)(void *))
{
    unsigned short  hashInd;
    ELLLIST        **pgpd = (ELLLIST **) gpdPvt;
    ELLLIST	*plist;
    GPDENTRY    *pgpdNode;

    if (gpdPvt == NULL)
      return;
    hashInd = (unsigned short)hash(name);
    if (pgpd[hashInd] == NULL) return;
    plist=pgpd[hashInd];
    pgpdNode = (GPDENTRY *) ellFirst(plist);
    while(pgpdNode) {
	if (strcmp(name,(char *)pgpdNode->name) == 0) {
	  if ((strlen(namepvt) == 0) ||
	      (strcmp(namepvt,pgpdNode->namepvt)==0)) {
	    ellDelete(plist, (ELLNODE*)pgpdNode);
	    (*userfree)(pgpdNode->userPvt);
	    free((void *)pgpdNode);
	    return;
          }
	}
	pgpdNode = (GPDENTRY *) ellNext((ELLNODE*)pgpdNode);
    }
    return;
}

void gpdFreeMem(void * gpdPvt, void (*userfree)(void *))
{
    unsigned short  hashInd;
    ELLLIST          **pgpd = (ELLLIST **) gpdPvt;
    ELLLIST           *plist;
    GPDENTRY       *pgpdNode;
    GPDENTRY       *next;;
    
    if (pgpd == NULL) return;
    for (hashInd=0; hashInd<HASH_NO; hashInd++) {
	if(pgpd[hashInd] == NULL) continue;
	plist=pgpd[hashInd];
	pgpdNode = (GPDENTRY *) ellFirst(plist);
	while(pgpdNode) {
	    next = (GPDENTRY *) ellNext((ELLNODE*)pgpdNode);
	    ellDelete(plist,(ELLNODE*)pgpdNode);
	    (*userfree)(pgpdNode->userPvt);
	    free((void *)pgpdNode);
	    pgpdNode = next;
	}
	free((void *)pgpd[hashInd]);
    }
    free((void *)pgpd);
}

void gpdDump(void * gpdPvt)
{
    unsigned short  hashInd;
    ELLLIST          **pgpd = (ELLLIST **) gpdPvt;
    ELLLIST           *plist;
    GPDENTRY       *pgpdNode;
    int		number;
    
    if (pgpd == NULL) return;
    for (hashInd=0; hashInd<HASH_NO; hashInd++) {
	if(pgpd[hashInd] == NULL) continue;
	plist=pgpd[hashInd];
	pgpdNode = (GPDENTRY *) ellFirst(plist);
	printf("\n %3.3hd=%3.3d",hashInd,ellCount(plist));
	number=0;
	while(pgpdNode) {
	    printf(" %s %s",pgpdNode->name,pgpdNode->namepvt);
	    if(number++ ==2) {number=0;printf("\n        ");}
	    pgpdNode = (GPDENTRY *) ellNext((ELLNODE*)pgpdNode);
	}
    }
    printf("\n End of General Purpose Directory\n");
}

GPDENTRY *gpdNext(void *gpdPvt, int op)
{
  static unsigned short hashInd = 0;
  ELLLIST **pgpd = (ELLLIST **)gpdPvt;
  ELLLIST *plist;
  static GPDENTRY *pgpdNode = NULL;
 
  if (op == 0) {
    hashInd = 0;
    pgpdNode = NULL;
    return(NULL);
  }
  if (pgpdNode != NULL)
    pgpdNode = (GPDENTRY *)ellNext((ELLNODE *)pgpdNode);

  if (pgpdNode == NULL) {
    plist = NULL;
    while ((hashInd < HASH_NO) && ((plist = pgpd[hashInd++]) == NULL))
      ;
    if (plist == NULL && hashInd == HASH_NO) {
      hashInd = 0;
      pgpdNode = NULL;
      return(NULL);
    }
    pgpdNode = (GPDENTRY *)ellFirst(plist);
  }
  return(pgpdNode);
}  
