/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/**** SCCS ID: @(#)libdev.h	1.21 2/1/94 ****/
/*************************************************************************
 * FILE:      libdev.h
 * Author:    Claude Saunders
 * Mods:      login mm/dd/yy: description
 *************************************************************************/

#ifdef __cplusplus
#include <c_varieties.h>
#endif

#include <sys/param.h>
#include <cadef.h>
#include <db_access.h>
#ifdef SOLARIS
/* # include <sunmath.h>*/		/* for quiet_nan */
#endif
#include <math.h>
#include "dev.h"
#include "devconsts.h"

typedef struct devnode {      /******* local device definition struct ****/
  char dname[MAXNAMELENDEV];     /* device name */
  int type;                   /* ATOMIC or COMPOSITE */
  char msg[MAXNAMELENDEV];       /* message name */
  char pvname[MAXNAMELENDEV];    /* associated process variable */
  int dir;                    /* READ or WRITE */
  int nargs;                  /* number of associated arguments (0 or 1) */
  int datatype;               /* DBR_DOUBLE or DBR_STRING y*/
  char def[MAXNAMELENDEV];       /* default value to write if nargs is 0 */
  chid *pchid;                /* ptr to chid for pvname */
  struct devnode *next;       /* additional devices if type is COMPOSITE */
} DEVNODE, *PDEVNODE;

typedef char **devnamearr;
typedef DEVNODE **devstructarr;

/* Error return codes from public function in libdev */
#define DEV_NORMAL        0
#define DEV_NODEV         1
#define DEV_NOMSG         2
#define DEV_CALLFAIL      3
#define DEV_NOSERVER      4
#define DEV_BADCHID       5
#define DEV_BADTYPE       6
#define DEV_TIMEOUT       7
#define DEV_BADSEARCH     8
#define DEV_ATOMICONLY    9

/* PUBLIC functions */

#ifdef __STDC__
#ifndef __cplusplus
/* basic */
int devCount(char *dname, int *cnt, devnamearr *names);
int devSend(char *dname, char *msg, void *arg);
int devSendBlock(char *dname, char *msg, void *arg);

/* advanced */
int devMatch(char *pattern, devnamearr *matches);
int devInfo(char *dname, devstructarr *infoarr);
int devDirType(char *msg, devstructarr info, int *dir, int *type);
int devChid(char *dname, char *msg, chid **pchid);
int devError(int status);
int devErrorString(int status, char *buf);
int devTimeout(float timeout);
#endif
#endif

#ifdef __cplusplus
/* basic */
EXTERN_FUNCTION(int devCount, (char *dname, int *cnt, devnamearr *names));
EXTERN_FUNCTION(int devSend, (char *dname, char *msg, void *arg));
EXTERN_FUNCTION(int devSendBlock, (char *dname, char *msg, void *arg));

/* advanced */
EXTERN_FUNCTION(int devMatch, (char *pattern, devnamearr *matches));
EXTERN_FUNCTION(int devInfo, (char *dname, devstructarr *infoarr));
EXTERN_FUNCTION(int devDirType, (char *msg, devstructarr info, int *dir, int *type));
EXTERN_FUNCTION(int devChid, (char *dname, char *msg, chid **pchid));
EXTERN_FUNCTION(int devError, (int status));
EXTERN_FUNCTION(int devErrorString, (int status, char *buf));
EXTERN_FUNCTION(int devTimeout, (float timeout));
#endif

